#ifdef HAVE_SYS_CDIO_H
#include <sys/cdio.h>
#endif

#ifdef __linux__
#include <linux/cdrom.h>
#include <sys/ioctl.h>
#endif

#ifdef __FreeBSD__
#include <camlib.h>
#include <sys/cdio.h>
#include <sys/cdrio.h>
#endif

#ifndef __linux__
#define CDS_NO_INFO   0
#define CDS_AUDIO   100
#define CDS_DATA_1  101
#define CDS_DATA_2  102
#define CDS_XA_2_1  103
#define CDS_XA_2_2  104
#define CDS_MIXED   105
#endif

#ifndef INVALID_HANDLE
#define INVALID_HANDLE (GINT_TO_POINTER(-1))
#endif

typedef struct
{
  const char *name;
  gboolean can_write_cdr;
  gboolean can_write_cdrw;
  gboolean can_write_dvdr;
  gboolean can_write_dvdram;
} OGMDvdWhiteDrive;

/* For dvd_plus_rw_utils.cpp */
int get_mmc_profile      (int        fd);
int get_disc_status      (int        fd,
                          int        *empty,
                          int        *is_rewritable,
                          int        *is_blank);
int get_dvd_r_rw_profile (const char *name);
/*
int get_read_write_speed (int        fd,
                          int        *read_speed,
                          int        *write_speed);
*/

#ifdef __FreeBSD__

#define get_ioctl_handle_fd(x) (((struct cam_device *)x)->fd)

gpointer
open_ioctl_handle (const char *device)
{
  struct cam_device *cam;

  cam = cam_open_device (device, O_RDWR);

  return (cam ? (gpointer) cam : INVALID_HANDLE);
}

void
close_ioctl_handle (gpointer handle)
{
  cam_close_device ((struct cam_device *) handle);
}

#else

#define get_ioctl_handle_fd(x) (GPOINTER_TO_INT(x))

gpointer
open_ioctl_handle (const char *device)
{
  int fd;

  if ((fd = open (device, O_RDWR | O_EXCL | O_NONBLOCK, 0)) < 0
      && (fd = open (device, O_RDONLY | O_EXCL | O_NONBLOCK, 0)) < 0)
    return INVALID_HANDLE;

  return GINT_TO_POINTER (fd);
}

void
close_ioctl_handle (gpointer handle)
{
  close (GPOINTER_TO_INT (handle));
}

#endif

/* adapted from gnome-vfs */
static int
get_disc_type (const char *dev_path)
{
  int fd = 0;
  int type;
#if defined (HAVE_SYS_MNTTAB_H) /* Solaris */
  GString *new_dev_path;
  struct cdrom_tocentry entry;
  struct cdrom_tochdr header;

  /*
   * For ioctl call to work dev_path has to be /vol/dev/rdsk.
   * There has to be a nicer way to do this.
   */
  new_dev_path = g_string_new (dev_path);
  new_dev_path = g_string_insert_c (new_dev_path, 9, 'r');
  fd = open (new_dev_path->str, O_RDONLY | O_NONBLOCK);
  g_string_free (new_dev_path, TRUE);

  if (fd < 0)
    return -1;

  if (ioctl (fd, CDROMREADTOCHDR, &header) < 0)
    return -1;

  type = CDS_DATA_1;

  for (entry.cdte_track = 1; entry.cdte_track <= header.cdth_trk1; entry.cdte_track++)
  {
    entry.cdte_format = CDROM_LBA;
    if (ioctl (fd, CDROMREADTOCENTRY, &entry) == 0)
    {
      if (!(entry.cdte_ctrl & CDROM_DATA_TRACK))
      {
        type = CDS_AUDIO;
        break;
      }
    }
  }

  return type;
#elif defined(HAVE_SYS_MNTCTL_H)
  return CDS_NO_INFO;
#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__OpenBSD__) || defined(__NetBSD__) || defined(__DragonFly__)
  struct ioc_toc_header header;
#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__DragonFly__)
  struct ioc_read_toc_single_entry entry;
#else
  struct ioc_read_toc_entry entries;
  struct cd_toc_entry entry;
  int i;
#endif

#ifndef CDROM_DATA_TRACK
#define CDROM_DATA_TRACK 4
#endif

  fd = open (dev_path, O_RDONLY | O_NONBLOCK);
  if (fd < 0)
    return -1;

  if (ioctl (fd, CDIOREADTOCHEADER, &header) < 0)
  {
    close (fd);
    return -1;
  }

  type = CDS_DATA_1;
#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__DragonFly__)
  for (entry.track = header.starting_track; entry.track <= header.ending_track; entry.track++)
  {
    entry.address_format = CD_LBA_FORMAT;
    if (ioctl (fd, CDIOREADTOCENTRY, &entry) == 0)
    {
      if (!(entry.entry.control & CDROM_DATA_TRACK))
      {
        type = CDS_AUDIO;
        break;
      }
    }
  }

#else /* defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__DragonFly__) */
  entries.data_len = sizeof (entry);
  entries.data = &entry;
  for (i = header.starting_track; i <= header.ending_track; i++)
  {
    entries.starting_track = i;
    entries.address_format = CD_LBA_FORMAT;
    if (ioctl (fd, CDIOREADTOCENTRYS, &entries) == 0)
    {
      if (!(entry.control & CDROM_DATA_TRACK))
      {
        type = CDS_AUDIO;
        break;
      }
    }
  }

#endif /* defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__DragonFly__) */
  close (fd);
  return type;
#elif defined(__linux__)
  fd = open (dev_path, O_RDONLY | O_NONBLOCK);
  if (fd < 0)
    return -1;

  if (ioctl (fd, CDROM_DRIVE_STATUS, CDSL_CURRENT) != CDS_DISC_OK)
  {
    close (fd);
    fd = -1;
    return -1;
  }

  type = ioctl (fd, CDROM_DISC_STATUS, CDSL_CURRENT);
  close (fd);

  return type;
#else /* defined(__linux__) */
  return CDS_NO_INFO;
#endif /* defined(__linux__) */
}

static void
disc_type_get_info (int type, gboolean *is_blank, gboolean *has_data, gboolean *has_audio)
{
  gboolean _has_audio = FALSE;
  gboolean _has_data = FALSE;
  gboolean _is_blank = FALSE;

  switch (type)
  {
    case CDS_AUDIO: /* audio CD */
      _has_audio = TRUE;
      break;
    case CDS_MIXED: /* mixed mode CD */
      _has_audio = TRUE;
      _has_data = TRUE;
      break;
    case CDS_DATA_1: /* data CD */
    case CDS_DATA_2:
    case CDS_XA_2_1:
    case CDS_XA_2_2:
      _has_data = TRUE;
      break;
    case CDS_NO_INFO: /* blank or invalid CD */
      _is_blank = TRUE;
      break;
    default: /* should never see this */
      break;
  }

  if (is_blank)
    *is_blank = _is_blank;
  if (has_data)
    *has_data = _has_data;
  if (has_audio)
    *has_audio = _has_audio;
}

static gboolean
drive_door_is_open (int fd)
{
  if (fd < 0)
    return FALSE;

#ifdef __linux__
  {
    int status;

    status = ioctl (fd, CDROM_DRIVE_STATUS, CDSL_CURRENT);
    if (status < 0)
      return FALSE;

    return status == CDS_TRAY_OPEN;
  }
#else
  return FALSE;
#endif
}

static gint
ogmdvd_drive_get_media_type_from_path (const gchar *device)
{
  gpointer ioctl_handle;
  int mmc_profile;
  int disc_type;
  int fd;

  g_return_val_if_fail (device != NULL, OGMDVD_MEDIA_TYPE_ERROR);

  ioctl_handle = open_ioctl_handle (device);
  if (ioctl_handle == INVALID_HANDLE)
  {
    if (errno == EBUSY)
      return OGMDVD_MEDIA_TYPE_BUSY;

    g_warning ("Could not open %s: %s", device, g_strerror (errno));

    return OGMDVD_MEDIA_TYPE_ERROR;
  }

  fd = get_ioctl_handle_fd (ioctl_handle);
  mmc_profile = get_mmc_profile (fd);

  /*
   * Couldn't get the data about the media 
   */
  if (mmc_profile < 0)
  {
    gboolean door_is_open;

    door_is_open = drive_door_is_open (fd);

    if (door_is_open != FALSE)
    {
      close_ioctl_handle (ioctl_handle);
      return OGMDVD_MEDIA_TYPE_ERROR;
    }
    else
    {
      int blank, rewrite, empty;

      if (get_disc_status (fd, &empty, &rewrite, &blank) == 0)
      {
        close_ioctl_handle (ioctl_handle);

        disc_type = get_disc_type (device);
        disc_type_get_info (disc_type, NULL, NULL, NULL);

        if (empty)
          return OGMDVD_MEDIA_TYPE_ERROR;
        else
          return OGMDVD_MEDIA_TYPE_UNKNOWN;
      }

      close_ioctl_handle (ioctl_handle);

      return OGMDVD_MEDIA_TYPE_UNKNOWN;
    }
  }

  close_ioctl_handle (ioctl_handle);

  disc_type = get_disc_type (device);
  disc_type_get_info (disc_type, NULL, NULL, NULL);

  switch (mmc_profile & 0xFFFF)
  {
    case -1:
      g_assert_not_reached ();
    case 0: /* No Media */
      return OGMDVD_MEDIA_TYPE_ERROR;
    case 0x8: /* Commercial CDs and Audio CD */
      return OGMDVD_MEDIA_TYPE_CD;
    case 0x9: /* CD-R */
      return OGMDVD_MEDIA_TYPE_CDR;
    case 0xa: /* CD-RW */
      return OGMDVD_MEDIA_TYPE_CDRW;
    case 0x10: /* Commercial DVDs */
      return OGMDVD_MEDIA_TYPE_DVD;
    case 0x11: /* DVD-R */
      return OGMDVD_MEDIA_TYPE_DVDR;
    case 0x13: /* DVD-RW Restricted Overwrite */
    case 0x14: /* DVD-RW Sequential */
      return OGMDVD_MEDIA_TYPE_DVDRW;
    case 0x1B: /* DVD+R */
      return OGMDVD_MEDIA_TYPE_DVD_PLUS_R;
    case 0x2B: /* DVD+R Double Layer */
      return OGMDVD_MEDIA_TYPE_DVD_PLUS_R_DL;
    case 0x1A: /* DVD+RW */
      return OGMDVD_MEDIA_TYPE_DVD_PLUS_RW;
    case 0x12: /* DVD-RAM  */
      return OGMDVD_MEDIA_TYPE_DVD_RAM;
    default:
      return OGMDVD_MEDIA_TYPE_UNKNOWN;
  }
}

void
ogmdvd_drive_add_whitelist (OGMDvdDrive *drive)
{
  OGMDvdWhiteDrive recorder_whitelist[] =
  {
    { "IOMEGA - CDRW9602EXT-B", TRUE, TRUE,  FALSE, FALSE },
    { "SONY - CD-R CDU948S",    TRUE, FALSE, FALSE, FALSE },
  };

  guint i;

  for (i = 0; i < G_N_ELEMENTS (recorder_whitelist); i++)
  {
    if (!strcmp (drive->priv->display_name, recorder_whitelist[i].name))
    {
      if (recorder_whitelist[i].can_write_cdr)
        drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_CD_RECORDER;
      if (recorder_whitelist[i].can_write_cdrw)
        drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_CDRW_RECORDER;
      if (recorder_whitelist[i].can_write_dvdr)
        drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_RW_RECORDER;
      if (recorder_whitelist[i].can_write_dvdram)
        drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_RAM_RECORDER;
    }
  }
}

static void
ogmdvd_drive_add_dvd_plus (OGMDvdDrive * drive)
{
  int caps;

  caps = get_dvd_r_rw_profile (drive->priv->device);

  if (caps == -1)
    return;

  if (caps == 2)
  {
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_PLUS_RW_RECORDER;
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_PLUS_R_RECORDER;
  }
  else if (caps == 0)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_PLUS_R_RECORDER;
  else if (caps == 1)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_PLUS_RW_RECORDER;
}

#if !defined (__linux__) && !defined (__FreeBSD__)

static char *
cdrecord_get_stdout_for_id (char *id)
{
  int max_speed, i;
  const char *argv[20];         /* Shouldn't need more than 20 arguments */
  char *dev_str, *stdout_data;

  max_speed = -1;

  i = 0;
  argv[i++] = "cdrecord";
  argv[i++] = "-prcap";
  dev_str = g_strdup_printf ("dev=%s", id);
  argv[i++] = dev_str;
  argv[i++] = NULL;

  if (g_spawn_sync (NULL, (char **) argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
        NULL, NULL, &stdout_data, NULL, NULL, NULL))
  {
    g_free (dev_str);
    return stdout_data;
  }

  g_free (dev_str);
  return NULL;
}

static char *
cdrecord_scan_get_stdout (void)
{
  int max_speed, i;
  const char *argv[20]; /* Shouldn't need more than 20 arguments */
  char *stdout_data;

  max_speed = -1;

  i = 0;
  argv[i++] = "cdrecord";
  argv[i++] = "-scanbus";
  argv[i++] = NULL;

  if (g_spawn_sync (NULL, (char **) argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
        NULL, NULL, &stdout_data, NULL, NULL, NULL))
    return stdout_data;

  return NULL;
}

static void
ogmdvd_drive_get_cd_properties (OGMDvdDrive *drive)
{
  char *stdout_data, *drive_cap;
/*
  drive->max_speed_read = -1;
  drive->max_speed_write = -1;
*/
  drive->priv->drive_type = 0;
/*
  drive->max_speed_read = get_device_max_read_speed (drive->device);
  drive->max_speed_write = get_device_max_write_speed (drive->device);
*/
  stdout_data = cdrecord_get_stdout_for_id (drive->priv->cdrecord_id);
  if (stdout_data == NULL)
    return;

  drive_cap = strstr (stdout_data, "Does write DVD-RAM media");
  if (drive_cap != NULL)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_RAM_RECORDER;
  drive_cap = strstr (stdout_data, "Does read DVD-R media");
  if (drive_cap != NULL)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_RW_RECORDER;
  drive_cap = strstr (stdout_data, "Does read DVD-ROM media");
  if (drive_cap != NULL)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_DRIVE;
  drive_cap = strstr (stdout_data, "Does write CD-RW media");
  if (drive_cap != NULL)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_CDRW_RECORDER;
  drive_cap = strstr (stdout_data, "Does write CD-R media");
  if (drive_cap != NULL)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_CD_RECORDER;
  drive_cap = strstr (stdout_data, "Does read CD-R media");
  if (drive_cap != NULL)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_CD_DRIVE;
  g_free (stdout_data);
}

static GList *
cdrecord_scan (gboolean recorder_only)
{
  GList *drives_list;
  OGMDvdDrive *drive;
  char *stdout_data, **lines, vendor[9], model[17];
  int i, bus, id, lun, index;

  drives_list = NULL;

  stdout_data = cdrecord_scan_get_stdout ();
  if (stdout_data == NULL)
    return drives_list;

  lines = g_strsplit (stdout_data, "\n", 0);
  g_free (stdout_data);

  for (i = 0; lines[i] != NULL; i++)
  {
    if (sscanf (lines[i], "\t%d,%d,%d\t  %d) '%8c' '%16c'", &bus, &id, &lun, &index, vendor, model) != 6)
      continue;

    vendor[8] = '\0';
    model[16] = '\0';

    drive = ogmdvd_drive_new ();
    drive->priv->display_name = g_strdup_printf ("%s - %s", g_strstrip (vendor), g_strstrip (model));
    drive->priv->cdrecord_id = g_strdup_printf ("%d,%d,%d", bus, id, lun);
    /*
     * FIXME we don't have any way to guess the real device
     * from the info we get from CDRecord 
     */
    drive->priv->device = g_strdup_printf ("/dev/pg%d", index);
    ogmdvd_drive_get_cd_properties (drive);
    ogmdvd_drive_add_whitelist (drive);
    if (drive->priv->drive_type & OGMDVD_DRIVE_TYPE_CD_RECORDER      ||
        drive->priv->drive_type & OGMDVD_DRIVE_TYPE_CDRW_RECORDER    ||
        drive->priv->drive_type & OGMDVD_DRIVE_TYPE_DVD_RAM_RECORDER ||
        drive->priv->drive_type & OGMDVD_DRIVE_TYPE_DVD_RW_RECORDER  || !recorder_only)
    {
/*
      if (drive->max_speed_read == -1)
        drive->max_speed_read = DEFAULT_SPEED;
      if (drive->max_speed_write == -1)
        drive->max_speed_write = DEFAULT_SPEED;
*/
      if (drive->priv->drive_type & OGMDVD_DRIVE_TYPE_DVD_DRIVE)
        ogmdvd_drive_add_dvd_plus (drive);

      drives_list = g_list_append (drives_list, drive);
    }
    else
      g_object_unref (drive);
  }

  g_strfreev (lines);

  return drives_list;
}

#endif

