#include <libhal.h>

#ifdef USE_HAL5
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#endif

static LibHalContext *ctx = NULL;

#ifdef USE_HAL5
static dbus_bool_t
hal_mainloop_integration (DBusError *error)
{
  DBusConnection *dbus_connection;

  dbus_connection = dbus_bus_get (DBUS_BUS_SYSTEM, error);

  if (dbus_error_is_set (error))
    return FALSE;

  dbus_connection_set_exit_on_disconnect (dbus_connection, FALSE);
  dbus_connection_setup_with_g_main (dbus_connection, NULL);

  libhal_ctx_set_dbus_connection (ctx, dbus_connection);

  return TRUE;
}
#endif

static void
ogmdvd_drive_hal_init (OGMDvdDrive *drive)
{
  if (!ctx)
  {
#ifdef USE_HAL5
    DBusError error;

    if (!(ctx = libhal_ctx_new ()))
    {
      g_warning ("failed to initialize HAL!");
      return;
    }

    dbus_error_init (&error);
    if (!hal_mainloop_integration (&error))
    {
      g_warning ("hal_initialize failed: %s", error.message);
      dbus_error_free (&error);
      return;
    }

    if (! libhal_ctx_init (ctx, &error))
    {
      g_warning ("hal_initialize failed: %s", error.message);
      dbus_error_free (&error);
      libhal_ctx_free (ctx);
      return;
    }
#else
    LibHalFunctions hal_functions =
    {
      NULL, /* mainloop integration */
      NULL, /* device_added */
      NULL, /* device_removed */
      NULL, /* device_new_capability */
      NULL, /* device_lost_capability */
      NULL, /* property_modified */
      NULL, /* device_condition */
    };

    ctx = hal_initialize (&hal_functions, FALSE);
#endif
  }
}

static void
ogmdvd_drive_hal_shutdown (OGMDvdDrive *drive)
{
#ifdef USE_HAL5
  if (ctx)
  {
    DBusError error;

    dbus_error_init (&error);
    if (!libhal_ctx_shutdown (ctx, &error))
    {
      g_warning ("hal_shutdown failed: %s\n", error.message);
      dbus_error_free (&error);
      return;
    }

    if (!libhal_ctx_free (ctx))
      g_warning ("hal_shutdown failed - unable to free hal context\n");

    ctx = NULL;
  }
#endif
}

#define LIBHAL_PROP_EXTRACT_BEGIN \
  if (FALSE)
#define LIBHAL_PROP_EXTRACT_INT(_property_, _where_) \
  else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_INT32) \
    _where_ = libhal_psi_get_int (&it)
#define LIBHAL_PROP_EXTRACT_STRING(_property_, _where_) \
  else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_STRING) \
    _where_ = (libhal_psi_get_string (&it) != NULL && strlen (libhal_psi_get_string (&it)) > 0) ? strdup (libhal_psi_get_string (&it)) : NULL
#define LIBHAL_PROP_EXTRACT_BOOL(_property_, _where_) \
  else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_BOOLEAN) \
    _where_ = libhal_psi_get_bool (&it)
#define LIBHAL_PROP_EXTRACT_BOOL_BITFIELD(_property_, _where_, _field_) \
  else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_BOOLEAN) \
    _where_ |= libhal_psi_get_bool (&it) ? _field_ : 0
#define LIBHAL_PROP_EXTRACT_STRLIST(_property_, _where_) \
  else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_STRLIST) \
    _where_ = libhal_psi_get_strlist (&it)
#define LIBHAL_PROP_EXTRACT_END ;

#define GET_BOOL_PROP(x) (hal_device_property_exists (ctx, udi, x) && hal_device_get_property_bool (ctx, udi, x))

static OGMDvdDrive *
hal_drive_from_udi (const char *udi)
{
  OGMDvdDrive *drive;

#ifdef USE_HAL5
  LibHalPropertySet        *pset;
  LibHalPropertySetIterator it;
  DBusError                 error;

  char *raw_device = NULL;
  char *cam_device = NULL;
  char *cam_path = NULL;

  LIBHAL_CHECK_LIBHALCONTEXT (ctx, FALSE);

  dbus_error_init (&error);

  if ((pset = libhal_device_get_all_properties (ctx, udi, &error)) == NULL)
  {
    if (dbus_error_is_set (&error))
    {
      g_warning ("Could not get all properties: %s", error.message);
      dbus_error_free (&error);
    }

    return NULL;
  }

  drive = ogmdvd_drive_new ();
  drive->priv->drive_type = OGMDVD_DRIVE_TYPE_CD_DRIVE;

  for (libhal_psi_init (&it, pset); libhal_psi_has_more (&it); libhal_psi_next (&it))
  {
    int   type;
    char *key;

    type = libhal_psi_get_type (&it);
    key = libhal_psi_get_key (&it);

    LIBHAL_PROP_EXTRACT_BEGIN;

    LIBHAL_PROP_EXTRACT_STRING ("block.device", drive->priv->device);
    LIBHAL_PROP_EXTRACT_STRING ("storage.model", drive->priv->display_name);

    LIBHAL_PROP_EXTRACT_STRING ("block.solaris.raw_device", raw_device);
    LIBHAL_PROP_EXTRACT_STRING ("block.freebsd.atapi_cam_device", cam_device);
    LIBHAL_PROP_EXTRACT_STRING ("block.freebsd.cam_path", cam_path);

    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.cdr",        drive->priv->drive_type, OGMDVD_DRIVE_TYPE_CD_RECORDER);
    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.cdrw",       drive->priv->drive_type, OGMDVD_DRIVE_TYPE_CDRW_RECORDER);
    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.dvd",        drive->priv->drive_type, OGMDVD_DRIVE_TYPE_DVD_DRIVE);
    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.dvdplusr",   drive->priv->drive_type, OGMDVD_DRIVE_TYPE_DVD_PLUS_R_RECORDER);
    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.dvdplusrw",  drive->priv->drive_type, OGMDVD_DRIVE_TYPE_DVD_PLUS_RW_RECORDER);
    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.dvdplusrdl", drive->priv->drive_type, OGMDVD_DRIVE_TYPE_DVD_PLUS_R_DL_RECORDER);

    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.dvdr",       drive->priv->drive_type, OGMDVD_DRIVE_TYPE_DVD_RW_RECORDER);
    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.dvdrw",      drive->priv->drive_type, OGMDVD_DRIVE_TYPE_DVD_RW_RECORDER);
    LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.dvdram",     drive->priv->drive_type, OGMDVD_DRIVE_TYPE_DVD_RAM_RECORDER);

    LIBHAL_PROP_EXTRACT_END;
  }

  if (raw_device != NULL)
  {
    g_free (drive->priv->device);
    drive->priv->device = raw_device;
  }

  if (cam_device != NULL)
  {
    g_free (drive->priv->device);
    drive->priv->device = cam_device;
  }

  if (cam_path)
    drive->priv->cdrecord_id = cam_path;
  else
    drive->priv->cdrecord_id = g_strdup (drive->priv->device);

  drive->priv->drive_udi = g_strdup (udi);

  libhal_free_property_set (pset);
#else
  gboolean is_cdr;

  is_cdr = GET_BOOL_PROP ("storage.cdrom.cdr");

  drive = ogmdvd_drive_new ();
  drive->priv->drive_type = OGMDVD_DRIVE_TYPE_CD_DRIVE;
  if (is_cdr != FALSE)
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_CD_RECORDER;
  if (GET_BOOL_PROP ("storage.cdrom.cdrw"))
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_CDRW_RECORDER;
  if (GET_BOOL_PROP ("storage.cdrom.dvd"))
  {
    drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_DRIVE;

    if (GET_BOOL_PROP ("storage.cdrom.dvdram"))
      drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_RAM_RECORDER;
    if (GET_BOOL_PROP ("storage.cdrom.dvdr"))
      drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_RW_RECORDER;
    if (GET_BOOL_PROP ("storage.cdrom.dvd"))
      drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_DRIVE;
    if (GET_BOOL_PROP ("storage.cdrom.dvdplusr"))
      drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_PLUS_R_RECORDER;
    if (GET_BOOL_PROP ("storage.cdrom.dvdplusrw"))
      drive->priv->drive_type |= OGMDVD_DRIVE_TYPE_DVD_PLUS_RW_RECORDER;
  }

  drive->priv->drive_udi = g_strdup (udi);
  drive->priv->device = hal_device_get_property_string (ctx, udi, "block.device");
  drive->priv->cdrecord_id = g_strdup (drive->priv->device);

  drive->priv->display_name = hal_device_get_property_string (ctx, udi, "storage.model");
#endif

  if (!drive->priv->display_name)
    drive->priv->display_name = g_strdup_printf ("Unnamed Drive (%s)", drive->priv->device);

  return drive;
}

static GList *
ogmdvd_drive_hal_scan (void)
{
  GList *drives = NULL;
  int i, num_devices;
  char **device_names;

  if (!ctx)
    ogmdvd_drive_hal_init (NULL);

#ifdef USE_HAL5
  device_names = libhal_find_device_by_capability (ctx, "storage.cdrom", &num_devices, NULL);
#else
  device_names = hal_find_device_by_capability (ctx, "storage.cdrom", &num_devices);
#endif

  if (!device_names)
    return NULL;

  for (i = 0; i < num_devices; i++)
  {
    OGMDvdDrive *drive;

    drive = hal_drive_from_udi (device_names [i]);
    if (drive)
    {
      ogmdvd_drive_add_whitelist (drive);
      drives = g_list_prepend (drives, drive);
    }
  }

#ifdef USE_HAL5
  libhal_free_string_array (device_names);
#else
  hal_free_string_array (device_names);
#endif

  return g_list_reverse (drives);
}

#ifdef USE_HAL5
static gboolean
hal_udi_is_our_drive (OGMDvdDrive *drive, const char *udi)
{
  char *device;
  gboolean res = FALSE;

  if (!drive->priv->device)
    return FALSE;

  if (!libhal_device_property_exists (ctx, udi, "block.device", NULL))
    return FALSE;

  device = libhal_device_get_property_string (ctx, udi, "block.device", NULL);
  if (device && strcmp (device, drive->priv->device) == 0)
    res = TRUE;

  if (device)
    libhal_free_string (device);

  return res;
}

static char *
ogmdvd_drive_get_first_child_udi (OGMDvdDrive *drive)
{
  DBusError error;
  char **device_names;
  int num_devices = -1;
  char *udi = NULL;

  if (!ctx)
    return NULL;

  if (!drive->priv->drive_udi)
    return NULL;

  dbus_error_init (&error);
  device_names = libhal_manager_find_device_string_match (ctx,
      "info.parent", drive->priv->drive_udi, &num_devices, &error);

  if (dbus_error_is_set (&error))
  {
    g_warning ("%s\n", error.message);
    dbus_error_free (&error);
  }
  else if (num_devices > 0)
    udi = g_strdup (device_names [0]);

  libhal_free_string_array (device_names);

  return udi;
}

static void
hal_device_added (LibHalContext *ctx, const char *udi)
{
  OGMDvdDrive *drive;

  drive = libhal_ctx_get_user_data (ctx);

  g_return_if_fail (drive != NULL);
  g_return_if_fail (udi != NULL);

  if (hal_udi_is_our_drive (drive, udi))
  {
    g_free (drive->priv->monitor_udi);
    drive->priv->monitor_udi = g_strdup (udi);

    g_signal_emit_by_name (drive, "media-added");
  }
}

static void
hal_device_removed (LibHalContext *ctx, const char *udi)
{
  OGMDvdDrive *drive;

  drive = libhal_ctx_get_user_data (ctx);

  g_return_if_fail (drive != NULL);
  g_return_if_fail (udi != NULL);

  if (drive->priv->monitor_udi && strcmp (drive->priv->monitor_udi, udi) == 0)
  {
    g_free (drive->priv->monitor_udi);
    drive->priv->monitor_udi = NULL;

    g_signal_emit_by_name (drive, "media-removed");
  }
}

static void
ogmdvd_drive_hal_set_monitor (OGMDvdDrive *drive, gboolean enabled)
{
  if (enabled)
  {
    libhal_ctx_set_user_data (ctx, drive);
    libhal_ctx_set_device_added (ctx, hal_device_added);
    libhal_ctx_set_device_removed (ctx, hal_device_removed);

    g_free (drive->priv->monitor_udi);
    drive->priv->monitor_udi = ogmdvd_drive_get_first_child_udi (drive);
  }
  else
  {
    libhal_ctx_set_user_data (ctx, NULL);
    libhal_ctx_set_device_added (ctx, NULL);
    libhal_ctx_set_device_removed (ctx, NULL);
  }

  /* FIXME: check for devices already in the drive and emit signals for them? */
}

gint
ogmdvd_drive_hal_get_media_type (OGMDvdDrive *drive)
{
  gchar *media_udi, *hal_type;
  gint media_type;

  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), OGMDVD_MEDIA_TYPE_ERROR);

  if (!drive->priv->drive_udi)
    return ogmdvd_drive_get_media_type_from_path (drive->priv->device);;

  media_udi = ogmdvd_drive_get_first_child_udi (drive);
  if (!media_udi)
    return OGMDVD_MEDIA_TYPE_UNKNOWN;

  hal_type = libhal_device_get_property_string (ctx, media_udi, "volume.disc.type", NULL);
  if (!hal_type)
    return OGMDVD_MEDIA_TYPE_UNKNOWN;

  if (strcmp (hal_type, "cd_rom") == 0)
    media_type = OGMDVD_MEDIA_TYPE_CD;
  else if (strcmp (hal_type, "cd_r") == 0)
    media_type = OGMDVD_MEDIA_TYPE_CDR;
  else if (strcmp (hal_type, "cd_rw") == 0)
    media_type = OGMDVD_MEDIA_TYPE_CDRW;
  else if (strcmp (hal_type, "dvd_rom") == 0)
    media_type = OGMDVD_MEDIA_TYPE_DVD;
  else if (strcmp (hal_type, "dvd_r") == 0)
    media_type = OGMDVD_MEDIA_TYPE_DVDR;
  else if (strcmp (hal_type, "dvd_ram") == 0)
    media_type = OGMDVD_MEDIA_TYPE_DVD_RAM;
  else if (strcmp (hal_type, "dvd_rw") == 0)
    media_type = OGMDVD_MEDIA_TYPE_DVDRW;
  else if (strcmp (hal_type, "dvd_plus_rw") == 0)
    media_type = OGMDVD_MEDIA_TYPE_DVD_PLUS_RW;
  else if (strcmp (hal_type, "dvd_plus_r") == 0)
    media_type = OGMDVD_MEDIA_TYPE_DVD_PLUS_R;
  else if (strcmp (hal_type, "dvd_plus_r_dl") == 0)
    media_type = OGMDVD_MEDIA_TYPE_DVD_PLUS_R_DL;
  else
    media_type = OGMDVD_MEDIA_TYPE_UNKNOWN;

  if (hal_type)
    libhal_free_string (hal_type);

  return media_type;
}
#endif

