/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-
 *
 * nautilus-burn-drive.c: easy to use cd burner software
 *
 * Copyright (C) 2002-2004 Red Hat, Inc.
 * Copyright (C) 2005 William Jon McCann <mccann@jhu.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 * Authors: Alexander Larsson <alexl@redhat.com>
 *          Bastien Nocera <hadess@hadess.net>
 *          William Jon McCann <mccann@jhu.edu>
 */

/**
 * SECTION:ogmdvd-drive
 * @title: OGMDvdDrive
 * @include: ogmdvd-drive.h
 * @short_description: An object representing an Optical Drive
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "ogmdvd-drive.h"

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

enum 
{
  MEDIA_ADDED,
  MEDIA_REMOVED,
  LAST_SIGNAL
};

enum 
{
  PROP_0,
  PROP_MONITOR_ENABLED
};

struct _OGMDvdDrivePriv
{
  gchar *device;
  gchar *cdrecord_id;
  gchar *display_name;
  gchar *drive_udi;
  gchar *monitor_udi;

  gint     monitor_poll_id;
  gboolean monitor_enabled;
  gboolean monitor_tray_open;
  gint     monitor_media_type;

  OGMDvdDriveType drive_type;
};

#include "ogmdvd-drive-common.c"

#ifdef __linux__
#include "ogmdvd-drive-linux.c"
#endif

#ifdef __FreeBSD__
#include "ogmdvd-drive-freebsd.c"
#endif

#if defined(USE_HAL4) || defined(USE_HAL5)
#include "ogmdvd-drive-hal.c"
#endif

#define OGMDVD_DRIVE_GET_PRIV(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), OGMDVD_TYPE_DRIVE, OGMDvdDrivePriv))

static void ogmdvd_drive_finalize     (GObject      *object);
static void ogmdvd_drive_set_property (GObject      *object,
                                       guint        prop_id,
                                       const GValue *value,
                                       GParamSpec   *pspec);
static void ogmdvd_drive_get_property (GObject      *object,
                                       guint        prop_id,
                                       GValue       *value,
                                       GParamSpec   *pspec);

static int signals [LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE (OGMDvdDrive, ogmdvd_drive, G_TYPE_OBJECT);

static void
ogmdvd_drive_class_init (OGMDvdDriveClass *klass)
{
  GObjectClass *object_class;

  object_class = (GObjectClass *) klass;

  object_class->finalize     = ogmdvd_drive_finalize;
  object_class->get_property = ogmdvd_drive_get_property;
  object_class->set_property = ogmdvd_drive_set_property;

  g_object_class_install_property (object_class, PROP_MONITOR_ENABLED, 
      g_param_spec_boolean ("enable-monitor",
                            NULL, NULL, FALSE,
                            G_PARAM_READWRITE));

  signals [MEDIA_ADDED] =
    g_signal_new ("media-added",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST, 
                  G_STRUCT_OFFSET (OGMDvdDriveClass,
                                   media_added),
                  NULL, NULL,
                  g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);

  signals [MEDIA_REMOVED] =
    g_signal_new ("media-removed",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST, 
                  G_STRUCT_OFFSET (OGMDvdDriveClass,
                                   media_removed),
                  NULL, NULL, 
                  g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);

  g_type_class_add_private (klass, sizeof (OGMDvdDrivePriv));
}

static void
ogmdvd_drive_init (OGMDvdDrive *drive)
{
  drive->priv = OGMDVD_DRIVE_GET_PRIV (drive);

#if defined(USE_HAL4) || defined(USE_HAL5)
  ogmdvd_drive_hal_init (drive);
#endif
}

static void
ogmdvd_drive_finalize (GObject *object)
{
  OGMDvdDrive *drive = OGMDVD_DRIVE (object);

  g_return_if_fail (object != NULL);

#if defined(USE_HAL4) || defined(USE_HAL5)
  ogmdvd_drive_hal_shutdown (drive);
#endif

  if (drive->priv->monitor_poll_id > 0)
    g_source_remove (drive->priv->monitor_poll_id);

  g_free (drive->priv->drive_udi);
  g_free (drive->priv->monitor_udi);

  g_free (drive->priv->display_name);
  g_free (drive->priv->cdrecord_id);
  g_free (drive->priv->device);

  if (G_OBJECT_CLASS (ogmdvd_drive_parent_class)->finalize != NULL)
    (* G_OBJECT_CLASS (ogmdvd_drive_parent_class)->finalize) (object);
}

static void
ogmdvd_drive_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  switch (prop_id) 
  {
    case PROP_MONITOR_ENABLED:
      ogmdvd_drive_set_monitor_enabled (OGMDVD_DRIVE (object), g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
ogmdvd_drive_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  switch (prop_id) 
  {
    case PROP_MONITOR_ENABLED:
      g_value_set_boolean (value, OGMDVD_DRIVE (object)->priv->monitor_enabled);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

#ifndef USE_HAL5
static gboolean
ogmdvd_drive_poll_tray_opened (OGMDvdDrive *drive)
{
  gboolean is_open;

  is_open = ogmdvd_drive_door_is_open (drive);

  /* check for no change */
  if (is_open == drive->priv->monitor_tray_open)
    return TRUE;

  if (is_open)
  {
    /* see if we lost media */
    if (drive->priv->monitor_media_type != OGMDVD_MEDIA_TYPE_ERROR)
    {
      drive->priv->monitor_media_type = OGMDVD_MEDIA_TYPE_ERROR;
      g_signal_emit_by_name (drive, "media-removed");
    }
  }
  else
  {
    OGMDvdDriveType type;

    /* see if there is now media in the drive */
    type = ogmdvd_drive_get_media_type (drive);

    if (type != OGMDVD_MEDIA_TYPE_ERROR)
    {
      drive->priv->monitor_media_type = type;
      g_signal_emit_by_name (drive, "media-added");
    }
  }

  drive->priv->monitor_tray_open = is_open;

  return TRUE;
}

static void
ogmdvd_drive_set_monitor (OGMDvdDrive *drive, gboolean enabled)
{
  if (drive->priv->monitor_poll_id > 0)
    g_source_remove (drive->priv->monitor_poll_id);

  if (enabled)
  {
    drive->priv->monitor_media_type = ogmdvd_drive_get_media_type (drive);
    drive->priv->monitor_tray_open = FALSE;
    drive->priv->monitor_poll_id = g_timeout_add (2000,
        (GSourceFunc) ogmdvd_drive_poll_tray_opened, drive);
  }
  else
  {
    drive->priv->monitor_media_type = OGMDVD_MEDIA_TYPE_ERROR;
    drive->priv->monitor_tray_open = FALSE;
    drive->priv->monitor_poll_id = 0;
  }
}
#endif

/**
 * ogmdvd_drive_new:
 *
 * Create a new #OGMDvdDrive.
 *
 * Return value: The new drive.
 **/
OGMDvdDrive *
ogmdvd_drive_new (void)
{
  return g_object_new (OGMDVD_TYPE_DRIVE, NULL);
}

/**
 * ogmdvd_drive_new_from_path:
 * @device: device path
 *
 * Creates a #OGMDvdDrive that corresponds to device.
 *
 * Return value: A new drive.
 **/
OGMDvdDrive *
ogmdvd_drive_new_from_path (const gchar *device)
{
  OGMDvdDrive *drive = NULL;
  GList *list, *link;

  list = ogmdvd_drive_get_list ();
  for (link = list; link; link = link->next) 
  {
    drive = link->data;
    if (g_str_equal (device, drive->priv->device)) 
      break;
  }

  if (!drive)
    return NULL;

  g_object_ref (drive);

  return drive;
}

/**
 * ogmdvd_drive_get_drive_type:
 * @drive: #OGMDvdDrive
 *
 * Get the type of the drive.
 *
 * Returns: type of drive
 **/
gint
ogmdvd_drive_get_drive_type (OGMDvdDrive *drive)
{
  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), 0);

  return drive->priv->drive_type;
}

/**
 * ogmdvd_drive_get_media_type:
 * @drive: #OGMDvdDrive
 *
 * Determine the type of the media in the drive @drive.
 *
 * Return value: The #OGMDvdMediaType of the media in the drive or the
 * following special values:
 *
 *    %OGMDVD_MEDIA_TYPE_ERROR   if the type can not be determined
 *    %OGMDVD_MEDIA_TYPE_UNKNOWN if the type can not be determined
 *    %OGMDVD_MEDIA_TYPE_BUSY    if the device is busy
 **/
gint
ogmdvd_drive_get_media_type (OGMDvdDrive *drive)
{
  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), OGMDVD_MEDIA_TYPE_ERROR);

#ifdef USE_HAL5
  return ogmdvd_drive_hal_get_media_type (drive);
#else
  return ogmdvd_drive_get_media_type_from_path (drive->priv->device);
#endif

  return 0;
}

/**
 * ogmdvd_drive_get_display_name:
 * @drive: #OGMDvdDrive
 *
 * Get the name of the drive for use in a user interface
 *
 * Returns: name of the drive.  Must be freed with g_free().
 **/
gchar *
ogmdvd_drive_get_display_name (OGMDvdDrive *drive)
{
  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), NULL);

  if (!drive->priv->display_name)
    return NULL;

  return g_strdup (drive->priv->display_name);
}

/**
 * ogmdvd_drive_get_device:
 * @drive: #OGMDvdDrive
 *
 * Get the name of the device associated with the drive.
 *
 * Returns: device name.  Must be not be freed.
 **/
const gchar *
ogmdvd_drive_get_device (OGMDvdDrive *drive)
{
  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), NULL);

  return drive->priv->device;
}

/**
 * ogmdvd_drive_eject:
 * @drive: #OGMDvdDrive
 *
 * Eject media from a #OGMDvdDrive.
 *
 * Return value: %TRUE if the media was sucessfully ejected, %FALSE otherwise.
 **/
gboolean
ogmdvd_drive_eject (OGMDvdDrive *drive)
{
  gchar *cmd;
  gboolean res;

  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), FALSE);

  if (drive->priv->device == NULL)
    return FALSE;

#ifdef __FreeBSD_
  cmd = g_strdup_printf ("cdcontrol -f %s eject", drive->priv->device);
#else
  cmd = g_strdup_printf ("eject %s", drive->priv->device);
#endif

  res = g_spawn_command_line_sync (cmd, NULL, NULL, NULL, NULL);
  g_free (cmd);

  /* delay a bit to make sure eject finishes */
  sleep (1);

  return res;
}

/**
 * ogmdvd_drive_door_is_open:
 * @drive: #OGMDvdDrive
 *
 * Report the whether the drive door or tray is open.
 *
 * Returns: %TRUE if the drive door is open, otherwise return %FALSE.
 **/
gboolean
ogmdvd_drive_door_is_open (OGMDvdDrive *drive)
{
  gpointer ioctl_handle;
  int      fd;
  gboolean ret;

  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), FALSE);

  ioctl_handle = open_ioctl_handle (drive->priv->device);
  if (ioctl_handle == INVALID_HANDLE)
    return FALSE;

  fd = get_ioctl_handle_fd (ioctl_handle);

  ret = drive_door_is_open (fd);

  close_ioctl_handle (ioctl_handle);

  return ret;
}

/**
 * ogmdvd_drive_set_monitor_enabled:
 * @drive: #OGMDvdDrive
 * @enabled: %TRUE to enable monitoring
 *
 * Sets whether to monitor @drive.
 **/
void
ogmdvd_drive_set_monitor_enabled (OGMDvdDrive *drive, gboolean enabled)
{
  g_return_if_fail (OGMDVD_IS_DRIVE (drive));

  if (drive->priv->monitor_enabled != enabled) 
  {
    drive->priv->monitor_enabled = enabled;

#ifdef USE_HAL5
    ogmdvd_drive_hal_set_monitor (drive, enabled);
#else
    ogmdvd_drive_set_monitor (drive, enabled);
#endif
  }
}

/**
 * ogmdvd_drive_get_monitor_enabled:
 * @drive: #OGMDvdDrive
 *
 * Gets whether the drive is monitored.
 *
 * Returns: %TRUE if @drive is monitored.
 **/
gboolean
ogmdvd_drive_get_monitor_enabled (OGMDvdDrive *drive)
{
  g_return_val_if_fail (OGMDVD_IS_DRIVE (drive), FALSE);

  return drive->priv->monitor_enabled;
}

/**
 * ogmdvd_drive_get_list:
 *
 * Get the list of available #OGMDvdDrive
 *
 * Return value: list of drives
 **/
GList *
ogmdvd_drive_get_list (void)
{
  static GList *drives = NULL;

  if (drives)
    return drives;

#if defined(USE_HAL4) || defined(USE_HAL5)
  drives = ogmdvd_drive_hal_scan ();
#endif

  if (drives == NULL)
  {
#if defined (__linux__)
    drives = linux_scan (FALSE);
#elif defined (__FreeBSD__)
    drives = freebsd_scan (FALSE);
#else
    drives = cdrecord_scan (FALSE);
#endif
  }

  return drives;
}

