#include "transport.h"

#include <poll.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#ifndef CHECK_CONDITION
#define CHECK_CONDITION 0x01
#endif

#ifndef EMEDIUMTYPE
#define EMEDIUMTYPE EINVAL
#endif

#ifndef	ENOMEDIUM
#define	ENOMEDIUM ENODEV
#endif

#define ERRCODE(s)      ((((s)[2]&0x0F)<<16)|((s)[12]<<8)|((s)[13]))
#define SK(errcode)     (((errcode)>>16)&0xF)
#define ASC(errcode)    (((errcode)>>8)&0xFF)
#define ASCQ(errcode)   ((errcode)&0xFF)

#define CREAM_ON_ERRNO_NAKED(s)                    \
    switch ((s)[12])                               \
    {   case 0x04: errno=EAGAIN;   break;          \
        case 0x20: errno=ENODEV;   break;          \
        case 0x21: if ((s)[13]==0) errno=ENOSPC;   \
                   else            errno=EINVAL;   \
                   break;                          \
        case 0x30: errno=EMEDIUMTYPE;      break;  \
        case 0x3A: errno=ENOMEDIUM;        break;  \
    }
#define CREAM_ON_ERRNO(s) do { CREAM_ON_ERRNO_NAKED(s) } while(0)

inline long
getmsecs ()
{
  struct timeval tv;
  gettimeofday (&tv, NULL);
  return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

void
sperror (const char *cmd, int err)
{
  int saved_errno = errno;

  if (err == -1)
    fprintf (stderr, ":-( unable to %s: ", cmd);
  else
    fprintf (stderr, ":-[ %s failed with SK=%Xh/ASC=%02Xh/ACQ=%02Xh]: ",
             cmd, SK (err), ASC (err), ASCQ (err));
  errno = saved_errno, perror (NULL);
}

#if defined(__linux)

#include <limits.h>
#include <mntent.h>

#include <linux/cdrom.h>

#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>

#include <scsi/sg.h>

#ifdef SG_IO
static const int Dir_xlate[] =
{
  SG_DXFER_NONE,
  SG_DXFER_FROM_DEV,
  SG_DXFER_TO_DEV
};

static int
use_sg_io (void)
{
  struct utsname buf;
  uname (&buf);
  /* was CDROM_SEND_PACKET declared dead in 2.5? */
  return (strcmp (buf.release, "2.5.43") >= 0);
}
#endif /* SG_IO */

struct _Scsi_Command
{
  int fd, autoclose;
  char *filename;
  struct cdrom_generic_command cgc;
  union
  {
    struct request_sense s;
    unsigned char u[18];
  } _sense;
#ifdef SG_IO
  struct sg_io_hdr sg_io;
#else /* SG_IO */
  struct
  {
    int cmd_len, timeout;
  } sg_io;
#endif /* SG_IO */
};

Scsi_Command *
scsi_command_new (void)
{
  Scsi_Command *cmd;

  cmd = (Scsi_Command *) malloc (sizeof (Scsi_Command));
  cmd->fd = -1;
  cmd->autoclose = 1;
  cmd->filename = NULL;

  return cmd;
}

Scsi_Command *
scsi_command_new_from_fd (int fd)
{
  Scsi_Command *cmd;

  cmd = (Scsi_Command *) malloc (sizeof (Scsi_Command));
  cmd->fd = fd;
  cmd->autoclose = 0;
  cmd->filename = NULL;

  return cmd;
}

void
scsi_command_free (Scsi_Command *cmd)
{
  if (cmd->fd >= 0 && cmd->autoclose)
  {
    close (cmd->fd);
    cmd->fd = -1;
  }

  if (cmd->filename)
  {
    free (cmd->filename);
    cmd->filename = NULL;
  }

  free (cmd);
}

int
openexcl (const char *device, int mode)
{
  int f, i;

  f = open (device, mode | O_EXCL);

  if (f < 0)
  {
    if (errno != EBUSY)
      return f;

    for (i = 0; (i < 10) && (f == -1); i++)
    {
      fprintf (stderr,
               "Error trying to open %s exclusively ... retrying in 1 second.\n",
               device);
      usleep (1000000 + (long) (100000.0 * rand () / (RAND_MAX + 1.0)));
      f = open (device, mode | O_EXCL);
    }

    if (f < 0)
    {
      fprintf (stderr, "Trying to open without locking.\n");
      f = open (device, mode);
    }
  }

  return f;
}

int
scsi_command_associate (Scsi_Command *cmd, const char *file)
{
  struct stat sb;

  /*
   * O_RDWR is expected to provide for none set-root-uid
   * execution under Linux kernel 2.6[.8]. Under 2.4 it
   * falls down to O_RDONLY...
   */
  if ((cmd->fd = openexcl (file, O_RDWR | O_NONBLOCK)) < 0 &&
      (cmd->fd = openexcl (file, O_RDONLY | O_NONBLOCK)) < 0)
    return 0;

  if (fstat (cmd->fd, &sb) < 0)
    return 0;
  if (!S_ISBLK (sb.st_mode))
  {
    errno = ENOTBLK;
    return 0;
  }

  cmd->filename = strdup (file);

  return 1;
}

void
scsi_command_set (Scsi_Command *cmd, size_t index, unsigned char value)
{
  if (index == 0)
  {
    memset (&cmd->cgc, 0, sizeof (cmd->cgc)), memset (&cmd->_sense, 0, sizeof (cmd->_sense));
    cmd->cgc.quiet = 1;
    cmd->cgc.sense = &cmd->_sense.s;
#ifdef SG_IO
    if (use_sg_io ())
    {
      memset (&cmd->sg_io, 0, sizeof (cmd->sg_io));
      cmd->sg_io.interface_id = 'S';
      cmd->sg_io.mx_sb_len = sizeof (cmd->_sense);
      cmd->sg_io.cmdp = cmd->cgc.cmd;
      cmd->sg_io.sbp = cmd->_sense.u;
      cmd->sg_io.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
    }
#endif /* SG_IO */
  }
  cmd->sg_io.cmd_len = index + 1;
  cmd->cgc.cmd[index] = value;
}

unsigned char *
scsi_command_sense (Scsi_Command *cmd)
{
  return cmd->_sense.u;
}

void
scsi_command_timeout (Scsi_Command *cmd, int i)
{
  cmd->cgc.timeout = cmd->sg_io.timeout = i * 1000;
}

#ifdef SG_IO
size_t
scsi_command_residue (Scsi_Command *cmd)
{
  return use_sg_io () ? cmd->sg_io.resid : 0;
}
#else /* SG_IO */
size_t
scsi_command_residue (Scsi_Command *cmd)
{
  return 0;
}
#endif /* SG_IO */

static const int real_dir[] =
{
  CGC_DATA_NONE,
  CGC_DATA_READ,
  CGC_DATA_WRITE
};

int
scsi_command_transport (Scsi_Command *cmd, Direction dir, void *buf, size_t sz)
{
  int ret = 0;

#ifdef SG_IO
#define KERNEL_BROKEN 0
  if (use_sg_io ())
  {
    cmd->sg_io.dxferp = buf;
    cmd->sg_io.dxfer_len = sz;
    cmd->sg_io.dxfer_direction = Dir_xlate[dir];
    if (ioctl (cmd->fd, SG_IO, &cmd->sg_io))
      return -1;

#if !KERNEL_BROKEN
    if ((cmd->sg_io.info & SG_INFO_OK_MASK) != SG_INFO_OK)
#else /* !KERNEL_BROKEN */
    if (cmd->sg_io.status)
#endif /* !KERNEL_BROKEN */
    {
      errno = EIO;
      ret = -1;
#if !KERNEL_BROKEN
      if (cmd->sg_io.masked_status & CHECK_CONDITION)
#endif /* KERNEL_BROKEN */
      {
        ret = ERRCODE ((unsigned char *) cmd->sg_io.sbp);
        if (ret == 0)
          ret = -1;
        else
          CREAM_ON_ERRNO ((unsigned char *) cmd->sg_io.sbp);
      }
    }
    return ret;
  }
  else
#undef KERNEL_BROKEN
#endif /* SG_IO */
  {
    cmd->cgc.buffer = (unsigned char *) buf;
    cmd->cgc.buflen = sz;
    cmd->cgc.data_direction = real_dir[dir];
    if (ioctl (cmd->fd, CDROM_SEND_PACKET, &cmd->cgc))
    {
      ret = ERRCODE (cmd->_sense.u);
      if (ret == 0)
        ret = -1;
    }
  }
  return ret;
}

int
scsi_command_umount (Scsi_Command *cmd, int f)
{
  struct stat fsb, msb;
  struct mntent *mb;
  FILE *fp;
  pid_t pid, rpid;
  int ret = 0, rval;

  if (f == -1)
    f = cmd->fd;
  if (fstat (f, &fsb) < 0)
    return -1;
  if ((fp = setmntent ("/proc/mounts", "r")) == NULL)
    return -1;

  while ((mb = getmntent (fp)) != NULL)
  {
    if (stat (mb->mnt_fsname, &msb) < 0)
      continue;               /* corrupted line? */
    if (msb.st_rdev == fsb.st_rdev)
    {
      ret = -1;
      if ((pid = fork ()) == (pid_t) - 1)
        break;
      if (pid == 0)
        execl ("/bin/umount", "umount", mb->mnt_dir, NULL);
      while (1)
      {
        rpid = waitpid (pid, &rval, 0);
        if (rpid == (pid_t) - 1)
        {
          if (errno == EINTR)
            continue;
          else
            break;
        }
        else if (rpid != pid)
        {
          errno = ECHILD;
          break;
        }
        if (WIFEXITED (rval))
        {
          if (WEXITSTATUS (rval) == 0)
            ret = 0;
          else
            errno = EBUSY;    /* most likely */
          break;
        }
        else
        {
          errno = ENOLINK;    /* some phony errno */
          break;
        }
      }
      break;
    }
  }
  endmntent (fp);

  return ret;
}

int
scsi_command_is_reload_needed (Scsi_Command *cmd)
{
  return ioctl (cmd->fd, CDROM_MEDIA_CHANGED, CDSL_CURRENT) == 0;
}

#elif defined(__sun) || defined(sun)

/* Licensing terms for commercial distribution for Solaris are to be
   settled with Inserve Technology, Åvägen 6, 412 50 GÖTEBORG, Sweden,
   tel. +46-(0)31-86 87 88, see http://www.inserve.se/ for further
  details. */

#include <volmgt.h>

int _dev_unmount (char *); /* VolMgt ON Consolidation Private API */

#include <sys/cdio.h>
#include <sys/dkio.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <sys/mnttab.h>
#include <sys/utsname.h>
#include <sys/dklabel.h>

#include <sys/scsi/impl/uscsi.h>

struct _Scsi_Command
{
  int fd;
  int autoclose;
  char *filename;
  struct uscsi_cmd ucmd;
  unsigned char cdb[16];
  unsigned char _sense[18];
};

Scsi_Command *
scsi_command_new (void)
{
  Scsi_Command *cmd;

  cmd = (Scsi_Command *) malloc (sizeof (Scsi_Command));
  cmd->fd = -1;
  cmd->autoclose = 1;
  cmd->filename = NULL;

  return cmd;
} 

Scsi_Command *
scsi_command_new_from_fd (int f)
{
  Scsi_Command *cmd;

  cmd = (Scsi_Command *) malloc (sizeof (Scsi_Command));
  cmd->fd = f;
  cmd->autoclose = 0;
  cmd->filename = NULL;

  return cmd;
}

void
scsi_command_free (Scsi_Command *cmd)
{
 if (cmd->fd >= 0 && cmd->autoclose)
 {
   close (cmd->fd);
   cmd->fd = -1;
 }

 if (cmd->filename)
 {
   free (cmd->filename);
   cmd->filename = NULL;
 }

 free (cmd);
}

int
scsi_command_associate (Scsi_Command *cmd, const char *file)
{
  char *volname, *device;
  struct stat sb;
  int v;

  if ((v = volmgt_running ()))
  {
    if ((volname = volmgt_symname ((char *) file)))
    {
      device = media_findname (volname);
      free (volname);

      if (!device)
        return 0;
    }
    else if ((device = media_findname ((char *) file)) == NULL)
      return 0;
  }
  else
    device = strdup (file);

  cmd->fd = open (device, O_RDONLY | O_NONBLOCK);
  if (cmd->fd < 0)
  {
    free (device);
    return 0;
  }
  if (fstat (cmd->fd, &sb) < 0)
  {
    free (device);
    return 0;
  }
  if (!S_ISCHR (sb.st_mode))
  {
    free (device);
    errno = ENOTTY;
    return 0;
  }

  cmd->filename = strdup (device);
  free (device);

  return 1;
}

void
scsi_command_set (Scsi_Command *cmd, size_t index, unsigned char value)
{
  if (index == 0)
  {
    memset (&cmd->ucmd, 0, sizeof (cmd->ucmd));
    memset (cmd->cdb, 0, sizeof (cmd->cdb));
    memset (cmd->_sense, 0, sizeof (cmd->_sense));
    cmd->ucmd.uscsi_cdb = (caddr_t) cmd->cdb;
    cmd->ucmd.uscsi_rqbuf = (caddr_t) cmd->_sense;
    cmd->ucmd.uscsi_rqlen = sizeof (cmd->_sense);
    cmd->ucmd.uscsi_flags = USCSI_SILENT | USCSI_DIAGNOSE |
      USCSI_ISOLATE | USCSI_RQENABLE;
    cmd->ucmd.uscsi_timeout = 60;
  }
  cmd->ucmd.uscsi_cdblen = index + 1;
  cmd->cdb[index] = value;
}

unsigned char *
scsi_command_sense (Scsi_Command *cmd)
{
  return cmd->_sense;
}

void
scsi_command_timeout (Scsi_Command *cmd, int i)
{
  cmd->ucmd.uscsi_timeout = i;
}

size_t
residue (Scsi_Command *cmd)
{
  return cmd->ucmd.uscsi_resid;
}

static const int real_dir[] =
{
  0,
  USCSI_READ,
  USCSI_WRITE
};

int
scsi_command_transport (Scsi_Command *cmd, Direction dir, void *buf, size_t sz)
{
  int ret = 0;

  cmd->ucmd.uscsi_bufaddr = (caddr_t) buf;
  cmd->ucmd.uscsi_buflen = sz;
  cmd->ucmd.uscsi_flags |= dir;
  if (ioctl (cmd->fd, USCSICMD, &cmd->ucmd))
  {
    if (errno == EIO && cmd->_sense[0] == 0) /* USB seems to be broken... */
    {
      size_t residue = cmd->ucmd.uscsi_resid;
      memset (cmd->cdb, 0, sizeof (cmd->cdb));
      cmd->cdb[0] = 0x03;          /* REQUEST SENSE */
      cmd->cdb[4] = sizeof (cmd->_sense);
      cmd->ucmd.uscsi_cdblen = 6;
      cmd->ucmd.uscsi_bufaddr = (caddr_t) cmd->_sense;
      cmd->ucmd.uscsi_buflen = sizeof (cmd->_sense);
      cmd->ucmd.uscsi_flags = USCSI_SILENT | USCSI_DIAGNOSE |
        USCSI_ISOLATE | USCSI_READ;
      cmd->ucmd.uscsi_timeout = 1;
      ret = ioctl (cmd->fd, USCSICMD, &cmd->ucmd);
      cmd->ucmd.uscsi_resid = residue;
      if (ret)
        return -1;
    }
    ret = ERRCODE (cmd->_sense);
    if (ret == 0)
      ret = -1;
    /*else  CREAM_ON_ERRNO(_sense); */
  }
  return ret;
}

/* mimics umount(2), therefore inconsistent return values */
int
scsi_command_umount (Scsi_Command *cmd, int f)
{
  struct stat fsb, msb;
  struct mnttab mb;
  FILE *fp;
  pid_t pid, rpid;
  int ret = 0, i, rval;

  if (f == -1)
    f = cmd->fd;

  if (fstat (f, &fsb) < 0)
    return -1;
  if ((fp = fopen (MNTTAB, "r")) == NULL)
    return -1;

  while ((i = getmntent (fp, &mb)) != -1)
  {
    if (i != 0)
      continue;               /* ignore corrupted lines */
    if (stat (mb.mnt_special, &msb) < 0)
      continue;               /* also corrupted?  */
    if (msb.st_rdev == fsb.st_rdev)
    {
      if (_dev_unmount (mb.mnt_special))
        break;
      {
        struct utsname uts;
        if (uname (&uts) >= 0 && strcmp (uts.release, "5.8") >= 0)
        {                     /* M-m-m-m-m! Solaris 8 or later... */
          ret = umount (mb.mnt_special);
          break;
        }
      }
      ret = -1;
      if ((pid = fork ()) == (pid_t) - 1)
        break;
      if (pid == 0)
        execl ("/usr/sbin/umount", "umount", mb.mnt_mountp, NULL);
      while (1)
      {
        rpid = waitpid (pid, &rval, 0);
        if (rpid == (pid_t) - 1)
        {
          if (errno == EINTR)
            continue;
          else
            break;
        }
        else if (rpid != pid)
        {
          errno = ECHILD;
          break;
        }
        if (WIFEXITED (rval))
        {
          if (WEXITSTATUS (rval) == 0)
            ret = 0;
          else
            errno = EBUSY;    /* most likely */
          break;
        }
        else if (WIFSTOPPED (rval) || WIFCONTINUED (rval))
          continue;
        else
        {
          errno = ENOLINK;    /* some phony errno */
          break;
        }
      }
      break;
    }
  }
  fclose (fp);

  return ret;
}

int
scsi_command_is_reload_needed (Scsi_Command *cmd)
{
  struct dk_minfo mi;
  struct dk_allmap pt;

  if (ioctl (cmd->fd, DKIOCGMEDIAINFO, &mi) < 0)
    return 1;

  memset (&pt, 0, sizeof (pt));
  pt.dka_map[2].dkl_nblk = mi.dki_capacity * (mi.dki_lbsize / DEV_BSIZE);
  pt.dka_map[0] = pt.dka_map[2];
  if (ioctl (cmd->fd, DKIOCSAPART, &pt) < 0)
    return 1;

  return 0;
}

#elif defined(__FreeBSD__)

#include <camlib.h>
#include <dirent.h>

#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/param.h>

#include <cam/scsi/scsi_message.h>
#include <cam/scsi/scsi_pass.h>

typedef off_t off64_t;
#define stat64   stat
#define fstat64  fstat
#define open64   open
#define pread64  pread
#define pwrite64 pwrite
#define lseek64  lseek

#define ioctl_fd (((struct cam_device *)ioctl_handle)->fd)

struct _Scsi_Command
{
  int fd;
  int autoclose;
  char *filename;
  struct cam_device *cam;
  union ccb ccb;
};

Scsi_Command *
scsi_command_new (void)
{
  Scsi_Command *cmd;

  cmd = (Scsi_Command *) malloc (sizeof (Scsi_Command));
  cmd->cam = NULL;
  cmd->fd = -1;
  cmd->autoclose = 1;
  cmd->filename = NULL;

  return cmd;
}

Scsi_Command *
scsi_command_new_from_fd (int f)
{
  Scsi_Command *cmd;

  char pass[32]; /* periph_name is 16 chars long */

  cmd = scsi_command_new ();

  memset (&cmd->ccb, 0, sizeof (cmd->ccb));
  cmd->ccb.ccb_h.func_code = XPT_GDEVLIST;
  if (ioctl (f, CAMGETPASSTHRU, &cmd->ccb) < 0)
  {
    free (cmd);
    return NULL;
  }

  sprintf (pass, "/dev/%.15s%u", cmd->ccb.cgdl.periph_name,
           cmd->ccb.cgdl.unit_number);
  cmd->cam = cam_open_pass (pass, O_RDWR, NULL);

  return cmd;
}

void
scsi_command_free (Scsi_Command *cmd)
{
  if (cmd->cam && cmd->autoclose)
  {
    cam_close_device (cmd->cam);
    cmd->cam = NULL;
  }

  if (cmd->fd >= 0)
    close (cmd->fd);

  if (cmd->filename)
  {
    free (cmd->filename);
    cmd->filename = NULL;
  }

  free (cmd);
}

int
scsi_command_associate (Scsi_Command *cmd, const char *file)
{
  struct stat sb;
  char pass[32]; /* periph_name is 16 chars long */

  cmd->fd = open (file, O_RDONLY | O_NONBLOCK);

  if (cmd->fd < 0)
    return 0;
  if (fstat (cmd->fd, &sb) < 0)
    return 0;
  if (!S_ISCHR (sb.st_mode))
    return (errno = EINVAL, 0);

  memset (&cmd->ccb, 0, sizeof (cmd->ccb));
  cmd->ccb.ccb_h.func_code = XPT_GDEVLIST;
  if (ioctl (cmd->fd, CAMGETPASSTHRU, &cmd->ccb) < 0)
  {
    close (cmd->fd);
    cmd->fd = -1;
    return 0;
  }

  sprintf (pass, "/dev/%.15s%u", cmd->ccb.cgdl.periph_name,
           cmd->ccb.cgdl.unit_number);
  cmd->cam = cam_open_pass (pass, O_RDWR, NULL);
  if (cmd->cam == NULL)
  {
    close (cmd->fd);
    cmd->fd = -1;
    return 0;
  }

  cmd->filename = strdup (file);

  return 1;
}

void
scsi_command_set (Scsi_Command *cmd, size_t index, unsigned char value)
{
  if (index == 0)
  {
    memset (&cmd->ccb, 0, sizeof (cmd->ccb));
    cmd->ccb.ccb_h.path_id = cmd->cam->path_id;
    cmd->ccb.ccb_h.target_id = cmd->cam->target_id;
    cmd->ccb.ccb_h.target_lun = cmd->cam->target_lun;
    cam_fill_csio (&(cmd->ccb.csio), 1,  /* retries */
                   NULL,      /* cbfncp */
                   CAM_DEV_QFRZDIS, /* flags */
                   MSG_SIMPLE_Q_TAG,  /* tag_action */
                   NULL,      /* data_ptr */
                   0,         /* dxfer_len */
                   sizeof (cmd->ccb.csio.sense_data),  /* sense_len */
                   0,         /* cdb_len */
                   30 * 1000);  /* timeout */
  }
  cmd->ccb.csio.cdb_len = index + 1;
  cmd->ccb.csio.cdb_io.cdb_bytes[index] = value;
}

unsigned char *
scsi_command_sense (Scsi_Command *cmd)
{
  return (unsigned char *) &cmd->ccb.csio.sense_data;
}

void
scsi_command_timeout (Scsi_Command *cmd, int i)
{
  cmd->ccb.ccb_h.timeout = i * 1000;
}

size_t
residue (Scsi_Command *cmd)
{
  return cmd->ccb.csio.resid;
}

static const int real_dir[] =
{
  CAM_DIR_NONE,
  CAM_DIR_IN,
  CAM_DIR_OUT
};

int
scsi_command_transport (Scsi_Command *cmd, Direction dir, void *buf, size_t sz)
{
  int ret = 0;

  cmd->ccb.csio.ccb_h.flags |= real_dir[dir];
  cmd->ccb.csio.data_ptr = (u_int8_t *) buf;
  cmd->ccb.csio.dxfer_len = sz;

  if ((ret = cam_send_ccb (cmd->cam, &cmd->ccb)) < 0)
    return -1;

  if ((cmd->ccb.ccb_h.status & CAM_STATUS_MASK) == CAM_REQ_CMP)
    return 0;

  unsigned char *sense = (unsigned char *) &cmd->ccb.csio.sense_data;

  errno = EIO;
  /* FreeBSD 5-CURRENT since 2003-08-24, including 5.2 fails to
     pull sense data automatically, at least for ATAPI transport,
     so I reach for it myself... */
  if ((cmd->ccb.csio.scsi_status == SCSI_STATUS_CHECK_COND) &&
      !(cmd->ccb.ccb_h.status & CAM_AUTOSNS_VALID))
  {
    u_int8_t _sense[18];
    u_int32_t resid = cmd->ccb.csio.resid;

    memset (_sense, 0, sizeof (_sense));

    scsi_command_set (cmd, 0, 0x03); /* REQUEST SENSE */
    cmd->ccb.csio.cdb_io.cdb_bytes[4] = sizeof (_sense);
    cmd->ccb.csio.cdb_len = 6;
    cmd->ccb.csio.ccb_h.flags |= CAM_DIR_IN | CAM_DIS_AUTOSENSE;
    cmd->ccb.csio.data_ptr = _sense;
    cmd->ccb.csio.dxfer_len = sizeof (_sense);
    cmd->ccb.csio.sense_len = 0;
    ret = cam_send_ccb (cmd->cam, &cmd->ccb);

    cmd->ccb.csio.resid = resid;
    if (ret < 0)
      return -1;
    if ((cmd->ccb.ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_CMP)
      return errno = EIO, -1;

    memcpy (sense, _sense, sizeof (_sense));
  }

  ret = ERRCODE (sense);
  if (ret == 0)
    ret = -1;
  else
    CREAM_ON_ERRNO (sense);

  return ret;
}

int
scsi_command_umount (Scsi_Command *cmd, int f)
{
  struct stat fsb, msb;
  struct statfs *mntbuf;
  int ret = 0, mntsize, i;

  if (f == -1)
    f = cmd->fd;

  if (fstat (f, &fsb) < 0)
    return -1;
  if ((mntsize = getmntinfo (&mntbuf, MNT_NOWAIT)) == 0)
    return -1;

  for (i = 0; i < mntsize; i++)
  {
    if (stat (mntbuf[i].f_mntfromname, &msb) < 0)
      continue;
    if (msb.st_rdev == fsb.st_rdev)
    {
      ret = unmount (mntbuf[i].f_mntonname, 0);
      break;
    }
  }

  return ret;
}

#define RELOAD_NEVER_NEEDED     /* according to Matthew Dillon */
int is_reload_needed ()
{
  return 0;
}

#endif

int
scsi_command_wait_for_unit (Scsi_Command *cmd, volatile int *progress)
{
  unsigned char *sense, sensebuf[18];
  int err;
  long msecs = 1000;

  sense = scsi_command_sense (cmd);

  while (1)
  {
    if (msecs > 0)
      poll (NULL, 0, msecs);
    msecs = getmsecs ();
    scsi_command_set (cmd, 0, 0); /* TEST UNIT READY */
    scsi_command_set (cmd, 5, 0);

    if (!(err = scsi_command_transport (cmd, NONE, NULL, 0)))
      break;
    /* I wish I could test just for sense.valid, but (at least)
       hp dvd100i returns 0 in valid bit at this point:-( So I
       check for all bits... */
    if ((sense[0] & 0x70) == 0)
    {
      perror (":-( unable to TEST UNIT READY");
      return -1;
    }
    else if (sense[12] == 0x3A) /* doesn't get any further than "no media" */
      return err;

    while (progress)
    {
      if (sense[15] & 0x80)
      {
        *progress = sense[16] << 8 | sense[17];
        break;
      }
      /* MMC-3 (draft) specification says that the unit should
         return progress indicator in key specific bytes even
         in reply to TEST UNIT READY. I.e. as above! But (at
         least) hp dvd100i doesn't do that and I have to fetch
         it separately:-( */
      scsi_command_set (cmd, 0, 0x03); /* REQUEST SENSE */
      scsi_command_set (cmd, 4, sizeof (sensebuf));
      scsi_command_set (cmd, 5, 0);

      if ((err = scsi_command_transport (cmd, READ, sensebuf, sizeof (sensebuf))))
      {
        sperror ("REQUEST SENSE", err);
        return err;
      }
      if (sensebuf[15] & 0x80)
        *progress = sensebuf[16] << 8 | sensebuf[17];
      break;
    }
    msecs = 1000 - (getmsecs () - msecs);
  }

  return 0;
}

